/* 
 * Comtrol RocketPort Infinity/Express Device Driver
 * FreeBSD 5 Operating System
 *
 * Copyright (c) 2008 Comtrol Corporation <support@comtrol.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the follwoing conditions
 * are met.
 * 1. Redistributions of source code must retain the above copyright 
 *    notive, this list of conditions and the following disclainer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials prodided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *       This product includes software developed by Comtrol Corporation.
 * 4. The name of Comtrol Corporation may not be used to endorse or 
 *    promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY COMTROL CORPORATION ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL COMTROL CORPORATION BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, LIFE OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
*/

/*
 *  For the device driver and tty methods, our softc is a CONTROLLER_T struct (CtlP).
 */
#include <sys/cdefs.h>

#include "opt_compat.h"

#include <sys/param.h>
#include <sys/proc.h>
#include <sys/systm.h>
#include <sys/fcntl.h>
#include <sys/malloc.h>
#include <sys/tty.h>
#include <sys/dkstat.h>
#include <sys/conf.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <machine/resource.h>
#include <machine/bus.h>
#include <sys/bus.h>
#include <sys/rman.h>
#include <sys/time.h>

#include <dev/rp_infinity/rp2.h>
#include <dev/rp_infinity/unity.h>
#include <dev/rp_infinity/unity.c>

#include <dev/pci/pcireg.h>
#include <dev/pci/pcivar.h>

static const char Rp2Version[] = "1.01";

/*
 *   The "driver" methods
 */
static int rp2_pciprobe(device_t dev);
static int rp2_pciattach(device_t dev);
static int rp2_pcidetach(device_t dev);
static int rp2_pcishutdown(device_t dev);

/*
 *   The "device" methods
 */
static int rp2open(struct cdev *dev, int flag, int mode, struct thread *td);
static int rp2close(struct cdev *dev, int flag, int mode, struct thread *td);
static int rp2write(struct cdev *dev, struct uio *uio, int flag);
static int rp2ioctl(struct cdev *dev, u_long cmd, caddr_t data, int flag, struct thread *td);

/*
 *  TTY Functions
 */
static int  rp2param(struct tty *, struct termios *);
static void rp2start(struct tty *);
static void rp2stop(struct tty *, int);


/*
 *   General driver functions
 */
static void rp2_do_poll(void *not_used);
static void rp2_read_poll(CHANNEL_T *ChP);
static void rp2_write_poll(CHANNEL_T *ChP);
static void rp2_untimeout(void);
static void rp2_pcireleaseresource(CONTROLLER_T *CtlP);
static int GetNumberOfModels(void);
static comtrolModelInfo_t *GetModelInfo(unsigned int model);
static void Rp2EnableInterrupts(CHANNEL_T *ChP, int modemLineInt);
static void Rp2DisableInterrupts(CHANNEL_T *ChP);
static int Rp2SetBaudRate(CHANNEL_T *ChP, int termiosBaud);
static void Rp2SetFlowControl(CHANNEL_T *ChP, int flowControl);
static CONTROLLER_T *Rp2GetControllerPtr(CHANNEL_T *ChP);
static void rp2hardclose(CHANNEL_T *ChP);

/*
 *   Declare the driver to the system
 */
static device_method_t rp2_pcimethods[] = {
  DEVMETHOD(device_probe, rp2_pciprobe),
  DEVMETHOD(device_attach, rp2_pciattach),
  DEVMETHOD(device_detach, rp2_pcidetach),
  DEVMETHOD(device_shutdown, rp2_pcishutdown),
  {0, 0}
};

static driver_t rp2_pcidriver = {
  "rp2",
  rp2_pcimethods,
  sizeof(CONTROLLER_T),
};

static devclass_t rp2_devclass;

DRIVER_MODULE(rp2, pci, rp2_pcidriver, rp2_devclass, 0, 0);

/*  Create the driver struct used in make_dev() in rp2_pciattach() */
struct cdevsw rp2_cdevsw = {
	.d_open = rp2open,
	.d_close = rp2close,
	.d_write = rp2write,
	.d_ioctl = rp2ioctl,
	.d_name = "rp2",
	.d_flags = D_TTY | D_NEEDGIANT,
	.d_version = D_VERSION,
};

/*  Static vars */
static int numberofBoards = 0;             /* Number of boards installed */ 
static int numPortsOpen = 0;               /* Number of open ports */
static int pollInterval;
struct callout_handle rp2_callout_handle;  /*  Used to start/stop the poll */
CONTROLLER_T *uController[MAX_NUMBEROF_BOARDS];  /*  Array of CONTROLLER_T pointers, so we can poll through the cards */

/*  Default prescale and corresponding baud divisor values */
#define DEFAULT_PRESCALE 0

/* The baudrate table - (termios enum, decimal baud, prescale, hardware divisor). */
baudTable_t defaultBaudTable[NUMBEROF_BAUDS] = {
    {B50,     50,      0, 55295},
    {B75,     75,      0, 36863},
    {B110,    110,     0, 25134},
    {B134,    134,     0, 20632},
    {B150,    150,     0, 18431},
    {B200,    200,     0, 13823},
    {B300,    300,     0, 9215},
    {B600,    600,     0, 4607},
    {B1200,   1200,    0, 2303},
    {B2400,   2400,    0, 1151},
    {B9600,   9600,    0, 287},
    {B19200,  19200,   0, 143},
    {B38400,  38400,   0, 71},
    {B57600,  57600,   0, 47},
    {B76800,  76800,   0, 35},
    {B115200, 115200,  0, 23},
    {B230400, 230400,  0, 11},
    {B460800, 460800,  0, 5},
    {B921600, 921600,  0, 2},
    {0, 0, 0}};


/******************************************/
/*             DRIVER METHODS             */
/******************************************/

/*
 *  Called once for each Comtrol card detected.  Verifies that this card is an Infinity
 *  using the PCI Device ID, as distinguished from Rocketport Device IDs.
 *  If it is an Infinity, passes success back to the system, which will call 
 *  rp2_pciattach().
 */
static int rp2_pciprobe(device_t dev)
{
    comtrolModelInfo_t *model;
    int i;
    struct timeval tv;

    DBGPRINT0("rp2_pciprobe() \n");

    if (pci_get_vendor(dev) != COMTROL_VENDOR_ID)
	return ENXIO;

    printf("Comtrol RocketPort Infinity(R) Driver - Release %s \n", Rp2Version);

    if (numberofBoards >= MAX_NUMBEROF_BOARDS) {
	device_printf(dev, "rp2_pciprobe() ERROR -Too many cards, max is %d \n", MAX_NUMBEROF_BOARDS);
	return ENXIO;
    }

    if (numberofBoards == 0) {
	for (i = 0; i < MAX_NUMBEROF_BOARDS; i++)
	    uController[i] = NULL;
    }

    /*  Retrieve the model info for the detected card */
    model = GetModelInfo(pci_get_device(dev));
    if (model == NULL) {

	//  This is not actually an error, other Comtrol RocketPort cards shares this 
	//  vendor ID, so this  probe function will get called if there is any of those cards installed.
//	device_printf(dev, "rckt_config() ERROR - model not found \n");
	return ENXIO;
    }

    numberofBoards++;
    device_set_desc(dev, model->modelString);

    /*  Set poll interval as close to 10ms as possible. */
    tv.tv_sec = 0;
    tv.tv_usec = 10000;
    pollInterval = tvtohz(&tv);
    if (pollInterval == 0)
	pollInterval = 1;
    callout_handle_init(&rp2_callout_handle);

    return BUS_PROBE_SPECIFIC;
}


/*
 *  Called once for each card if probe() was successful.
 *
 *  NOTE on per device structs.  Each device (card) has a CONTROLLER_T struct associated 
 *  with it.  In FreeBSD, this is created by the system, and passed to the driver 
 *  as the "softc".
 *
 *  Returns:  0 if successful
 */
static int rp2_pciattach(device_t dev)
{
    CONTROLLER_T *CtlP;
    CHANNEL_T *ChP;
    int ch;
    uint32_t pciCfg;

    DBGPRINT0("rp2_pciattach() \n");

    /*  Initialize the CONTROLLER_T struct associated with this card */
    CtlP = device_get_softc(dev);
    uController[numberofBoards - 1] = CtlP;
    memset(CtlP, 0, sizeof(CONTROLLER_T));
    for (ch = 0; ch < MAX_PORTS_PER_BOARD; ch++) {
	ChP = &CtlP->channel[ch];
	memset(ChP, 0, sizeof(CHANNEL_T));
	ChP->channelNumber = ch;
	ChP->portNumber = ch + ((numberofBoards - 1) * MAX_PORTS_PER_BOARD);
    }
    CtlP->dev = dev;

    /*  Enable the PCI part of the card, get the device ID, retrieve model info from our tables */
    if (pci_enable_io(dev, SYS_RES_MEMORY))
	return ENXIO;

    pciCfg = pci_read_config(dev, PCIR_COMMAND, 4);
    if (!(pciCfg & PCIM_CMD_MEMEN)) {
	device_printf(dev, "rp2_pciattach() ERROR - failed to enable memory mapping \n");
	return ENXIO;;
    }

    /*  Retrieve the model info for the detected card */
    CtlP->rocketModel = GetModelInfo(pci_get_device(dev));
    if (CtlP->rocketModel == NULL) {   /*  Should never happen, probe() should catch this... */
	device_printf(dev, "rp2_pciattach() Error - model not found \n");
	return ENXIO;
    }

    //  For PCI, the resource ID is the offset into pci config space which has 
    //  the word describing the resource.
    CtlP->rid0 = PCIR_BAR(0);
    CtlP->rid1 = PCIR_BAR(1);

    /*  Retrieve PCI base address region 0 */
    CtlP->pciResource0 = bus_alloc_resource(dev, SYS_RES_MEMORY, &CtlP->rid0, 0ul, ~0ul, 1, RF_ACTIVE);
    if (CtlP->pciResource0 == NULL) {
	device_printf(dev, "rp2_pciattach() ERROR - unable to allocate memory resource 0 \n");
	return ENXIO;;
    }
    CtlP->pBasePciRegion0 = rman_get_virtual(CtlP->pciResource0);

    /*  Retrieve PCI base address region 1 */
    CtlP->pciResource1 = bus_alloc_resource(dev, SYS_RES_MEMORY, &CtlP->rid1, 0ul, ~0ul, 1, RF_ACTIVE);
    if (CtlP->pciResource1 == NULL) {
	device_printf(dev, "rp2_pciattach() ERROR - unable to allocate memory resource 1 \n");
	return ENXIO;
    }
    CtlP->pBasePciMemory = rman_get_virtual(CtlP->pciResource1);

    CtlP->has_2nd_asic = CtlP->rocketModel->has_2nd_asic;

#ifdef  RCKT_DEBUG
    device_printf(dev, "Card detected model: 0x%x  str: %s  PCIBaseAddr: %p  UnityBase: %p  numUnityAsics: %d  numPorts: %d \n", 
		  (int)CtlP->rocketModel->model, CtlP->rocketModel->modelString, CtlP->pBasePciRegion0, CtlP->pBasePciMemory, (int)CtlP->has_2nd_asic + 1, (int)CtlP->rocketModel->numPorts);
#else
    printf("\nRP2: Comtrol RP Infinity/Express Card detected - Model: %s  Number of Ports: %d \n", CtlP->rocketModel->modelString, (int)CtlP->rocketModel->numPorts);
    printf("RP2: Ports can be accessed using: \n");
    
    switch (numberofBoards - 1) {
    case 0:
	printf("RP2: /dev/ttyRPI0  - /dev/ttyRPI%d  (normal serial ports) \n",   (int)CtlP->rocketModel->numPorts - 1);
	printf("RP2: /dev/cuaRPI0  - /dev/cuaRPI%d  (callout ports) \n",      (int)CtlP->rocketModel->numPorts - 1);
	break;
    case 1:
	printf("RP2: /dev/ttyRPI32  - /dev/ttyRPI%d  (normal serial ports) \n", ((int)CtlP->rocketModel->numPorts - 1) + 32);
	printf("RP2: /dev/cuaRPI32  - /dev/cuaRPI%d  (callout ports) \n",    ((int)CtlP->rocketModel->numPorts - 1) + 32);
	break;
    case 2:
	printf("RP2: /dev/ttyRPI64  - /dev/ttyRPI%d  (normal serial ports) \n", ((int)CtlP->rocketModel->numPorts - 1) + 64);
	printf("RP2: /dev/cuaRPI64  - /dev/cuaRPI%d  (callout ports) \n",    ((int)CtlP->rocketModel->numPorts - 1) + 64);
	break;
    case 3:
	printf("RP2: /dev/ttyRPI96  - /dev/ttyRPI%d  (normal serial ports) \n", ((int)CtlP->rocketModel->numPorts - 1) + 96);
	printf("RP2: /dev/cuaRPI96  - /dev/cuaRPI%d  (callout ports) \n",    ((int)CtlP->rocketModel->numPorts - 1) + 96);
	break;
    default: 
	break;
    }

#endif   //  ifdef RCKT_DEBUG

    /*  
     *  The Flextronics Chipset requires the following registers to be set, 
     *  else, all PCI memory accesses to region 1 will return garbage.
     */
    WRITEL(4, CtlP->pBasePciRegion0 + 0x110);
    WRITEL(0, CtlP->pBasePciRegion0 + 0x11C);

    uGlobalReset(CtlP);
    uInitAsicConfigReg(CtlP);
    uSetClockPrescale(CtlP, DEFAULT_PRESCALE);

    /*  Store the channel's memory address, each is offset by 0x40 from the base */
    for (ch = 0; ch < MAX_PORTS_PER_BOARD; ch++) {

	/*  If the port actually exists on the card, init it */
	if (ch < CtlP->rocketModel->numPorts) {
	    ChP = &CtlP->channel[ch];

	    ChP->rp_restart = 0;
	    ChP->rp_xmit_stopped = 0;
	    ChP->Present = 1;

	    /*  Make the ttyRPIxx and cuaRPIxx devices */
	    ChP->dev0 = make_dev(&rp2_cdevsw, 
				 ((numberofBoards - 1) << 16) | ChP->portNumber,        
				 UID_ROOT, 
				 GID_WHEEL,  
				 0644, 
				 "ttyRPI%d", 
				 ChP->portNumber);
	    ChP->dev1 = make_dev(&rp2_cdevsw, 
				 ((numberofBoards - 1) << 16) | ChP->portNumber | 0x80, 
				 UID_UUCP, 
				 GID_DIALER, 
				 0644, 
				 "cuaRPI%d", 
				 ChP->portNumber);

	    /*  Store the PCI memory addresses to access the port's hardware */
	    if (ch < 16) {
		ChP->pBaseRegister = CtlP->pBasePciMemory + (ch * 0x40);
		ChP->ucodeAddr     = CtlP->pBasePciMemory + UCODE_SRAM_CORE0 + (ch * 0x80);
	    }
	    else {
		ChP->pBaseRegister = CtlP->pBasePciMemory + 0x1000 + ((ch - 16) * 0x40);
		ChP->ucodeAddr     = CtlP->pBasePciMemory + 0x1000 + UCODE_SRAM_CORE0 + ((ch - 16) * 0x80);
	    }
	    
	    /*  Init hardware to good defaults */
	    Rp2SetFlowControl(ChP, FLOW_CONTROL_NONE);
	    uSetBaudDivisor(ChP, 287);             //  9600 baud
	    uSetData8(ChP);                        // 8N1 RS232
	    uDisParity(ChP);
	    uSetStop1(ChP);
	    if ((CtlP->rocketModel->model == RP2_MODEL_SMPTE_8) ||
		(CtlP->rocketModel->model == RP2_EXPRESS_MODEL_SMPTE_8))
		uSetInterfaceMode(ChP, URS_RS422);  
	    else
		uSetInterfaceMode(ChP, URS_RS232);  
	    uClrDtr(ChP);
	    uClrRts(ChP);
	    Rp2DisableInterrupts(ChP);

	    /*  Create the tty struct for the port, filled in during open() */
	    ChP->rp2_tty = ttymalloc(NULL);
	    ChP->rp2_tty->t_sc = ChP;

	    /*  Set the initial state termios vals for in and out */
	    ChP->it_in.c_iflag = 0;
	    ChP->it_in.c_oflag = 0;
	    ChP->it_in.c_cflag = TTYDEF_CFLAG;
	    ChP->it_in.c_lflag = 0;
	    termioschars(&ChP->it_in);
	    ChP->it_in.c_ispeed = ChP->it_in.c_ospeed = TTYDEF_SPEED;
	    ChP->it_out = ChP->it_in;
	}
    }

    return 0;
}

/*
 *
 */
static int rp2_pcidetach(device_t dev)
{
    CONTROLLER_T *CtlP;

    DBGPRINT0("rp2_pcidetach() \n");
    if (device_get_state(dev) == DS_BUSY)
	return (EBUSY);

    CtlP = device_get_softc(dev);
    rp2_pcireleaseresource(CtlP);
    return 0;
}

/*
 *
 */
static int rp2_pcishutdown(device_t dev)
{
    DBGPRINT0("rp2_pcishutdown() \n");
    return rp2_pcidetach(dev);
}

/*  
 *  Breaks out of any waiting states, then frees the PCI memory resources
 *  allocated during attach().
 */
static void rp2_pcireleaseresource(CONTROLLER_T *CtlP)
{
    DBGPRINT0("rp2_releaseresource() \n");
    rp2_untimeout();
    if (CtlP->pciResource0 != NULL) {
	bus_release_resource(CtlP->dev, SYS_RES_MEMORY, CtlP->rid0, CtlP->pciResource0);
	CtlP->pciResource0 = NULL;
    }
    if (CtlP->pciResource1 != NULL) {
	bus_release_resource(CtlP->dev, SYS_RES_MEMORY, CtlP->rid1, CtlP->pciResource1);
	CtlP->pciResource1 = NULL;
    }
}

/******************************************/
/*             DEVICE METHODS             */
/******************************************/

/*
 *  Device method that opens the port.  Note that each port has a CHANNEL_T associated 
 *  with it, stored in the device's softc struct.  This function initializes the port.
 *  If this is the first open port, polling is started.
 */
static int rp2open(struct cdev *dev, int flag, int mode, struct thread *td)
{
    int mynor, unit, portnum, oldspl, error;
    CONTROLLER_T *CtlP;
    CHANNEL_T *ChP;
    struct tty *tp;

    mynor = minor(dev);
    unit = MINOR_TO_UNIT(mynor);
    portnum = MINOR_TO_PORT(mynor);
    portnum -= (unit * 32);
    CtlP = (CONTROLLER_T *)devclass_get_softc(rp2_devclass, unit);
    ChP = &CtlP->channel[portnum];
    tp = ChP->rp2_tty;

    DBGPRINT0("rp2open() \n");

    if (!ChP->Present) {            /* chan found and init'd? */
	printf("RP2: ERROR in rp2open() - line not present, returning ENXIO \n");
        return(ENXIO);              /* no.. fail the open */
    }

    oldspl = spltty();

    dev->si_tty = tp;
    tp->t_dev = dev;
    tp->t_param = rp2param;
    tp->t_oproc = rp2start;
    tp->t_stop = rp2stop;
    tp->t_line = 0;
    tp->t_termios = IS_CALLOUT(dev) ? ChP->it_out : ChP->it_in;
    tp->t_ififosize = 512;
    tp->t_ispeedwat = (speed_t)-1;
    tp->t_ospeedwat = (speed_t)-1;

    /*  Load ucode, enable transmitter, set initial state */
    uSetRxTrigger(ChP, URCV_FIFO_TRIG_1_CHAR);
    uFlushRxFifo(ChP);           
    uFlushTxFifo(ChP);
    Rp2EnableInterrupts(ChP, 0);
    uSetXon(ChP); 
    uLoadUcode(ChP);
    uStartRxProcessor(ChP);
    uEnTransmit(ChP);
    uSetDtr(ChP);
    uSetRts(ChP);
    uEnSpecialReceiveInt(ChP); 

    rp2param(tp, &tp->t_termios);
		
    //  If this is the first port to open, start the poll
    numPortsOpen++;
    if (numPortsOpen == 1)
	rp2_callout_handle = timeout(rp2_do_poll, (void *) NULL, pollInterval);

    error = ttyld_open(tp, dev);
    ttyldoptim(tp);
    splx(oldspl);
    if (error == 0)
	device_busy(CtlP->dev);
    return (error);
}

/*
 *  Device method that closes the port.
 */
static int rp2close(struct cdev *dev, int flag, int mode, struct thread *td)
{
    int mynor, unit, portnum, oldspl;
    CONTROLLER_T *CtlP;
    CHANNEL_T *ChP;
    struct tty *tp;

    mynor = minor(dev);
    unit = MINOR_TO_UNIT(mynor);
    portnum = MINOR_TO_PORT(mynor);
    portnum -= (unit * 32);
    CtlP = (CONTROLLER_T *)devclass_get_softc(rp2_devclass, unit);
    ChP = &CtlP->channel[portnum];
    tp = ChP->rp2_tty;

    oldspl = spltty();
    ttyld_close(tp, flag);
    ttyldoptim(tp);
    rp2stop(tp, FREAD | FWRITE);
    rp2hardclose(ChP);
    tp->t_state &= ~TS_BUSY;
    tty_close(tp);
    splx(oldspl);
    device_unbusy(CtlP->dev);
    if (numPortsOpen > 0)
	numPortsOpen--;

    return 0;
}


/*
 *  Device method that closes the port.
 */
static int rp2write(struct cdev *dev, struct uio *uio, int flag)
{
    int mynor, unit, portnum;
    CONTROLLER_T *CtlP;
    CHANNEL_T *ChP;
    struct tty *tp;

    mynor = minor(dev);
    unit = MINOR_TO_UNIT(mynor);
    portnum = MINOR_TO_PORT(mynor);
    portnum -= (unit * 32);
    CtlP = (CONTROLLER_T *)devclass_get_softc(rp2_devclass, unit);
    ChP = &CtlP->channel[portnum];
    tp = ChP->rp2_tty;

    return ttyld_write(tp, uio, flag);
}


/*
 *  Device method that implements ioctls
 */
static int rp2ioctl(struct cdev *dev, u_long cmd, caddr_t data, int flag, struct thread *td)
{
    int mynor, unit, portnum, error, oldspl;
    CONTROLLER_T *CtlP;
    CHANNEL_T *ChP;
    struct tty *tp;
    struct termios *t;
    int *ModemMaskPtr;  /* arg for TIOCMSET */
    int ModemMask;      /* return for TIOCMGET */
    int *IfModePtr;     /* arg for RP2_IOCTL_SETIFMODE */

    mynor = minor(dev);
    unit = MINOR_TO_UNIT(mynor);
    portnum = MINOR_TO_PORT(mynor);
    portnum -= (unit * 32);
    CtlP = (CONTROLLER_T *)devclass_get_softc(rp2_devclass, unit);
    ChP = &CtlP->channel[portnum];
    tp = ChP->rp2_tty;

//    DBGPRINT1("RP2: rp2ioctl() - cmd: 0x%x \n", (int)cmd);
    switch(cmd) {

    case RP2_IOCTL_SETIFMODE:
	IfModePtr = (int *)data;
	if (IfModePtr[0] == IFMODE_RS232) {
	    DBGPRINT0("RP2:  Setting RS232 mode\n");
	    uSetInterfaceMode(ChP, URS_RS232);
	    uDisRTSToggle(ChP);
	}
	else if (IfModePtr[0] == IFMODE_RS422) {
	    DBGPRINT0("RP2:  Setting RS422 mode\n");
	    uSetInterfaceMode(ChP, URS_RS422);
	    uDisRTSToggle(ChP);
	}
	else if (IfModePtr[0] == IFMODE_RS485) {
	    DBGPRINT0("RP2:  Setting RS485 mode\n");
	    uSetInterfaceMode(ChP, URS_RS485_HD);
	    uEnRTSToggle(ChP);
	}
	else {
	    DBGPRINT1("IOC_IFMODE - invalid arg (0x%x)", IfModePtr[0]);
	}
	return (0);
	
    default:
	break;
    }

    t = &tp->t_termios;
    error = ttyioctl(dev, cmd, data, flag, td);
    ttyldoptim(tp);
    if (error != ENOTTY)
	return (error);

    oldspl = spltty();

    switch(cmd) {

    case TIOCSBRK:
	uSendBreak(ChP);
	break;

    case TIOCCBRK:
	uClrBreak(ChP);
	break;

    case TIOCSDTR:
	uSetDtr(ChP);
	uSetRts(ChP);
	break;
	
    case TIOCCDTR:
	uClrDtr(ChP);
	break;

    case TIOCMSET: 

       ModemMaskPtr = (int *)data;
       if (*ModemMaskPtr & TIOCM_DTR) 
	   uSetDtr(ChP);
       else 
	   uClrDtr(ChP);
       if (*ModemMaskPtr & TIOCM_RTS) 
	   uSetRts(ChP);
       else 
	   uClrRts(ChP);
       break;

    case TIOCMBIS:
	ModemMaskPtr = (int *)data;
	if (*ModemMaskPtr & TIOCM_DTR)
	    uSetDtr(ChP);
	if (*ModemMaskPtr & TIOCM_RTS)
	    uSetRts(ChP);
	break;

    case TIOCMBIC:
	ModemMaskPtr = (int *)data;
	if (*ModemMaskPtr & TIOCM_DTR)
	    uClrDtr(ChP);
	if (*ModemMaskPtr & TIOCM_RTS)
	    uClrRts(ChP);
	break;

    case TIOCMGET:
	ModemMask = 0;
	if (uCdActive(ChP))
	    ModemMask |= TIOCM_CAR;
	if (uDsrActive(ChP))
	    ModemMask |= TIOCM_DSR;
	if (uCtsActive(ChP))
	    ModemMask |= TIOCM_CTS;
	if (uDtrActive(ChP))
	    ModemMask |= TIOCM_DTR;
	if (uRtsActive(ChP))
	    ModemMask |= TIOCM_RTS;
	*(int *)data = ModemMask;
	break;

    default:
	DBGPRINT1("RP2: rp2ioctl() ERROR - Invalid IOCTL %d \n", (int)cmd);
	splx(oldspl);
	return ENOTTY;
    }

    splx(oldspl);
    return 0;
}



/*
 *  The main worker poll routine.  Polls all installed ports for reading/writing
 *  work to be done.  Runs every pollInterval if any ports are open.
 */
static void rp2_do_poll(void *not_used)
{
    CONTROLLER_T *CtlP;
    CHANNEL_T *ChP;
    static int firstTimeThrough = 1;
    int ctrl, ch;

    if (firstTimeThrough) {
	firstTimeThrough = 0;
	DBGPRINT0("RP2: rp2_do_poll() \n");
    }

    /*  Walk through each of the installed boards */
    for (ctrl = 0; ctrl < MAX_NUMBEROF_BOARDS; ctrl++) {
	CtlP = uController[ctrl];
	if (CtlP == NULL)   //  Card not installed, no more cards after this
	    break;

	/*  Walk through each of the installed ports */
	for (ch = 0; ch < CtlP->rocketModel->numPorts; ch++) {
	    ChP = &CtlP->channel[ch];
	    if (ChP->Present) {

		/*  The read side signals us via the "interrupts" - flags in the ASIC */
		if (uGetInterruptStatus(CtlP, ch)) 
		    rp2_read_poll(ChP);
		
		/*  The write side is polled to see if work needs to be done there */
		rp2_write_poll(ChP);
	    }
	}
    }

    /*  If there are still ports open, schedule the next poll */
    if (numPortsOpen > 0)
	rp2_callout_handle = timeout(rp2_do_poll, (void *) NULL, pollInterval);
}


/*
 * The read side of the port polling routine.  Work that is done here:
 * - If Rx data is present, read from the RX FIFO and put into buffer
 * - If Errors, handle them  
 * - If CD change, send appropriate messages. 
 */
static void rp2_read_poll(CHANNEL_T *ChP)
{
    struct tty *tp = ChP->rp2_tty;
    unsigned int Count; /*  The number of bytes in the RX FIFO */
    int ch;             /*  The "char (actually an int) received, with error flags attached */
    int spl;

//    DBGPRINT0("RP2: rp2_read_poll() \n");

    /*  Check for received serial data */
    if (uRcvFifoTriggerInt(ChP) || uRcvDataRdyInt(ChP)) {

	if (!(tp->t_state & TS_TBLOCK) && 
	    (tp->t_state & TS_CARR_ON) && 
	    (tp->t_state & TS_ISOPEN)) {

	    Count = uGetRxCnt(ChP);
	    while (Count) {
		if (tp->t_state & TS_TBLOCK)
		    break;

		/*  Get the byte, then add error flags if necessary */
		ch = (int)uRxByte(ChP);
	  
		/*  
		 * Special receive is set if there is a parity, framing, break, (there are more, 
		 * see Unity spec) error in the RCV FIFO.  This driver handles those 3, ignores rest. 
		 */
		if (uSpecialReceiveInt(ChP)) {
		    DBGPRINT0("RP2: rp2_read_poll() - special receive set \n");
		  
		    if (uChanStatusBreakDetected(ChP)) {
			DBGPRINT0("RP2: rp2_read_poll() - break detected \n");
			ch |= TTY_FE;
		    }
		  
		    if (uChanStatusFramingError(ChP)) {
			DBGPRINT0("RP2: rp2_read_poll() - framing error detected \n");
			ch |= TTY_FE;
		    }
		  
		    if (uChanStatusParityError(ChP)) {
			DBGPRINT0("RP2: rp2_read_poll() - parity error detected \n");
			ch |= TTY_PE;
		    }
		    uClrSpecialReceiveInt(ChP);
		}

		/*  Send byte+flags up the tty chain */
		spl = spltty();
		ttyld_rint(tp, ch);
		splx(spl);
		Count--;
	    }
	}
    }
	
    /*  Handle a change in the CD line */	
    if (uIntCdChange(ChP)) {
	DBGPRINT0("RP2: ttyr_poll() - CD change \n");
	
	/*  Is CD asserted? */
	if (uCdActive(ChP)) {
	    
	    if (!(tp->t_state & TS_CARR_ON))
		ttyld_modem(tp, 1);
	}
	else {
	    if (tp->t_state & TS_CARR_ON) {
		(void) ttyld_modem(tp, 0);
		if (ttyld_modem(tp, 0) == 0)
		    rp2hardclose(ChP);
	    }
	}
    }
}

/*
 * The write side of the port polling routine.  No actual work is done here, 
 * rp2_start() is triggered if there is room in the TX FIFO. 
 */
static void rp2_write_poll(CHANNEL_T *ChP)
{
    struct tty *tp = ChP->rp2_tty;
    int count;

//    DBGPRINT0("rp2_write_poll() \n");
    if ((tp->t_state & TS_BUSY) && (tp->t_state & TS_ISOPEN)) {
	count = uGetTxCnt(ChP);
	if (count == 0)
	    tp->t_state &= ~(TS_BUSY);
	if (!(tp->t_state & TS_TTSTOP) && (count <= ChP->rp_restart)) {
	    ttyld_start(tp);
	}
    }
}
    

/*
 *  Cancels the timed polling routine.
 */
static void rp2_untimeout(void)
{
    DBGPRINT0("rp2_untimeout() \n");
    untimeout(rp2_do_poll, (void *) NULL, rp2_callout_handle);
}

/*
 *  TTY Function - Set the port params (bit, baud, parity, etc) to the values in 
 *  the passed-in termios struct.  The values in the tty struct are updated.
 */
static int rp2param(struct tty *tp, struct termios *t)
{
    int mynor, unit, portnum, oldspl;
    int cflag, iflag, oflag, lflag, ret;
    CONTROLLER_T *CtlP;
    CHANNEL_T *ChP;

    mynor = minor(tp->t_dev);
    unit = MINOR_TO_UNIT(mynor);
    portnum = MINOR_TO_PORT(mynor);
    portnum -= (unit * 32);
    CtlP = (CONTROLLER_T *)devclass_get_softc(rp2_devclass, unit);
    ChP = &CtlP->channel[portnum];

    DBGPRINT0("rp2param() \n");

    oldspl = spltty();

    cflag = t->c_cflag;
    iflag = t->c_iflag;
    oflag = t->c_oflag;
    lflag = t->c_lflag;

    tp->t_ispeed = t->c_ispeed;
    tp->t_ospeed = t->c_ospeed;
    tp->t_cflag = cflag;
    tp->t_iflag = iflag;
    tp->t_oflag = oflag;
    tp->t_lflag = lflag;

    if (t->c_ospeed == 0) {
	uClrDtr(ChP);
	splx(oldspl);
	return (0);
    }
    
    ChP->rp_fifo_lw = ((t->c_ospeed * 2) / 1000) + 1;

    uSetDtr(ChP);
    uSetRts(ChP);
    ret = Rp2SetBaudRate(ChP, t->c_ospeed);
    if (ret == -1) 
	Rp2SetBaudRate(ChP, B9600);

    if ((cflag & CSIZE) == CS8)
	uSetData8(ChP);
    else    
	uSetData7(ChP);
	
    if (cflag & CSTOPB)
	uSetStop2(ChP); 
    else
	uSetStop1(ChP); 
	
    if (cflag & PARENB)
	uEnParity(ChP); 
    else
	uDisParity(ChP); 
	
    if (cflag & PARODD)
	uSetOddParity(ChP); 
    else
	uSetEvenParity(ChP); 

    /* ISTRIP strips the high bits from incoming characters */
    if (iflag & ISTRIP)
	uSetMask(ChP, 0x7f); 
    else 
	uSetMask(ChP, 0xff); 

    /*  If CLOCAL is set, disable the delta CD interrupt.  If clear, enable it. */ 
    if (cflag & CLOCAL)
	Rp2EnableInterrupts(ChP, 0);
    else
	Rp2EnableInterrupts(ChP, 1);

    if (cflag & CCTS_OFLOW) {
	DBGPRINT0("rp2param() - ENABLE CTS FLOW \n");
	uEnCtsFlowCtrl(ChP); 
    }
    else  {
	DBGPRINT0("rp2param() - DISABLE CTS FLOW \n");
	uDisCtsFlowCtrl(ChP); 
    }

    if (cflag & CRTS_IFLOW) {
	DBGPRINT0("rp2param() - ENABLE RTS FLOW \n");
	uEnRtsFlowCtrl(ChP); 
    }
    else {
	DBGPRINT0("rp2param() - DISABLE RTS FLOW \n");
	uDisRtsFlowCtrl(ChP);
	uSetRts(ChP);
    }

    ttyldoptim(tp);

    if ((cflag & CLOCAL) || (uCdActive(ChP))) {
	tp->t_state |= TS_CARR_ON;
	wakeup(TSA_CARR_ON(tp));
    }

    splx(oldspl);
    return 0;
}


/*
 *   TTY Function (oproc) - Drains the TTY xmit queue, and transmits the bytes
 */
static void rp2start(struct tty *tp)
{
    int mynor, unit, portnum, oldspl;
    int Count, xmit_fifo_room, i, numToXmit;
    CONTROLLER_T *CtlP;
    CHANNEL_T *ChP;
    struct clist *qp;

    mynor = minor(tp->t_dev);
    unit = MINOR_TO_UNIT(mynor);
    portnum = MINOR_TO_PORT(mynor);
    portnum -= (unit * 32);
    CtlP = (CONTROLLER_T *)devclass_get_softc(rp2_devclass, unit);
    ChP = &CtlP->channel[portnum];

    DBGPRINT0("rp2start() \n");
    oldspl = spltty();

    if (tp->t_state & (TS_TIMEOUT | TS_TTSTOP))  {
	ttwwakeup(tp);
	splx(oldspl);
	return;
    }

    if (ChP->rp_xmit_stopped) {
	uEnTransmit(ChP);
	ChP->rp_xmit_stopped = 0;
    }

    Count = uGetTxCnt(ChP);

    if (tp->t_outq.c_cc == 0)    {
	if ((tp->t_state & TS_BUSY) && (Count == 0))
	    tp->t_state &= ~TS_BUSY;
	ttwwakeup(tp);
	splx(oldspl);
	return;
    }

    xmit_fifo_room = FIFO_SIZE - Count;
    qp = &tp->t_outq;
    if (xmit_fifo_room > 0 && qp->c_cc > 0) {
	tp->t_state |= TS_BUSY;

	//  Push the bytes from the clist (*qp) into the TxBuf, then into the TX FIFO */
	numToXmit = q_to_b(qp, ChP->TxBuf, xmit_fifo_room);
	for (i = 0; i < numToXmit; i++)
	    uXmitByte(ChP, ChP->TxBuf[i]);
    }
    ChP->rp_restart = (qp->c_cc > 0) ? ChP->rp_fifo_lw : 0;

    ttwwakeup(tp);
    splx(oldspl);
}

/*
 *  TTY Function - Flushes ASIC TX FIFO, disables the transmitter.
 */
static void rp2stop(register struct tty *tp, int flag)
{
    int mynor, unit, portnum, oldspl;
    CONTROLLER_T *CtlP;
    CHANNEL_T *ChP;

    mynor = minor(tp->t_dev);
    unit = MINOR_TO_UNIT(mynor);
    portnum = MINOR_TO_PORT(mynor);
    portnum -= (unit * 32);
    CtlP = (CONTROLLER_T *)devclass_get_softc(rp2_devclass, unit);
    ChP = &CtlP->channel[portnum];

    DBGPRINT0("rp2stop() \n");
    oldspl = spltty();

    if (tp->t_state & TS_BUSY) {
	if ((tp->t_state & TS_TTSTOP) == 0) {
	    uFlushTxFifo(ChP);
	}
	else {
	    if (ChP->rp_xmit_stopped == 0) {
		uDisTransmit(ChP);
		ChP->rp_xmit_stopped = 1;
	    }
	}
    }
    splx(oldspl);
    rp2start(tp);
}


static void rp2hardclose(CHANNEL_T *ChP)
{
    struct tty *tp;
    
    tp = ChP->rp2_tty;
    
    uFlushRxFifo(ChP);           
    uFlushTxFifo(ChP);
    uDisTransmit(ChP);
    Rp2DisableInterrupts(ChP);
    Rp2SetFlowControl(ChP, FLOW_CONTROL_NONE);
    uClrDtr(ChP);
    uClrRts(ChP);
}

/***************************************************************
 *
 *  Rocketport Infinity models and accessor functions
 *
 ***************************************************************/

/*  Comtrol RocketPort Infinity Model Info Array - Every device and it's attributes */
static comtrolModelInfo_t comtrolModelInfo[] = {

        /*   model               model String                           numPorts   has_2nd_asic */
	{RP2_MODEL_INTF_32, "RocketPort Infinity External I/F 32 Port",  32,           1},
	{RP2_MODEL_INTF_8,  "RocketPort Infinity External I/F 8 Port",    8,           0},
	{RP2_MODEL_INTF_16, "RocketPort Infinity External I/F 16 Port",  16,           0},
	{RP2_MODEL_DB9_4,   "RocketPort Infinity DB9 4 Port Card",        4,           0},
	{RP2_MODEL_DB9_8,   "RocketPort Infinity DB9 8 Port Card",        8,           0},
	{RP2_MODEL_SMPTE_8, "RocketPort Infinity SMPTE 8 Port Card",      8,           0},

        /*   model                   model String                           numPorts   has_2nd_asic */
	{RP2_EXPRESS_MODEL_RJ45_8,  "RocketPort Express RJ45 8 Port Card",       8,           0},
	{RP2_EXPRESS_MODEL_INTF_32, "RocketPort Express External I/F 32 Port",  32,           1},
	{RP2_EXPRESS_MODEL_INTF_8,  "RocketPort Express External I/F 8 Port",    8,           0},
	{RP2_EXPRESS_MODEL_INTF_16, "RocketPort Express External I/F 16 Port",  16,           0},
	{RP2_EXPRESS_MODEL_DB9_4,   "RocketPort Express DB9 4 Port Card",        4,           0},
	{RP2_EXPRESS_MODEL_DB9_8,   "RocketPort Express DB9 8 Port Card",        8,           0},
	{RP2_EXPRESS_MODEL_INTF_4,  "RocketPort Express External I/F 4 Port",    4,           0},
	{RP2_EXPRESS_MODEL_SMPTE_8, "RocketPort Express SMPTE 8 Port Card",      8,           0}

}; 

static int GetNumberOfModels(void) 
{
    return (sizeof(comtrolModelInfo) / sizeof(comtrolModelInfo_t));
}

/* 
 *  Gets the comtrolModelInfo struct corresponding to the model type
 *  Inputs:   Model type (PCI Model ID)
 *  Returns:  comtrolModelInfo_t pointer, or NULL if not found
 */
static comtrolModelInfo_t *GetModelInfo(unsigned int model)
{
    int i;

    for (i = 0; i < GetNumberOfModels(); i++) {
	if (comtrolModelInfo[i].model ==  model)
	    return &comtrolModelInfo[i];
    }
    return NULL;
}


/***************************************************************************** 
 *  Enables/disables the standard interrupts used by the driver.  
 *  These are:
 *  -  TX FIFO Empty
 *  -  RX Trigger 
 *  -  Special Receive
 *  -  Modem Control Change
 *
 *  The channel is then unmasked/masked in the Channel Interrupt Mask.
 *
 *  Note that this does not enable hardware interrupts, interrupts are not used
 *  in this driver.  The interrupts enabled/disabled are  the interrupt sources.
 *  Enabling them allows them to be polled.
 *
 *   Seems only 2 interrupts have any affect, RX FIFO trigger and CD change
 *   Inputs: Channel ptr and flag indicating whether to enable modem line ints.
 *****************************************************************************/
static void Rp2EnableInterrupts(CHANNEL_T *ChP, int modemLineInt)
{
    CONTROLLER_T *CtlP = Rp2GetControllerPtr(ChP);
    if (CtlP == NULL) {
	printf("RP2:  Rp2EnableInterrupts ERROR - Unable to get controller ptr \n");
	return;
    }

    DBGPRINT0("Rp2EnableInterrupts() \n");

    uSetRxTrigger(ChP, URCV_FIFO_TRIG_1_CHAR);
    if (modemLineInt)
	uEnModemChangeInt(ChP);
    uEnChannelInterrupt(CtlP, ChP->channelNumber);
}

static void Rp2DisableInterrupts(CHANNEL_T *ChP)
{
    CONTROLLER_T *CtlP = Rp2GetControllerPtr(ChP);
    if (CtlP == NULL) {
	printf("RP2:  Rp2DisableInterrupts ERROR - Unable to get controller ptr \n");
	return;
    }

    DBGPRINT0("Rp2DisableInterrupts() \n");

    uSetRxTrigger(ChP,  URCV_FIFO_TRIG_DISABLED);
    uDisModemChangeInt(ChP);
    uDisChannelInterrupt(CtlP, ChP->channelNumber);
}

/*  
 *   Sets the baud divisor for the requested termios baud rate value (B9600, etc).
 *   Inputs:  Channel pointer and baud rate
 *   Returns:   0 if OK, -1 else
 */
static int Rp2SetBaudRate(CHANNEL_T *ChP, int termiosBaud)
{
    int i = 0, baudFound = 0, baudRate = 0;
    int baudDivisor = 0;
    
    DBGPRINT0("Rp2SetBaudRate() \n");

    /*  Find baud in table */
    for (i = 0; i < NUMBEROF_BAUDS; i++) {
	if (defaultBaudTable[i].termiosBaud == termiosBaud) {
	    baudFound = 1;
	    baudDivisor = defaultBaudTable[i].divisor;
	    baudRate = defaultBaudTable[i].baud;
	    break;
	}
    }
    if (!baudFound) {
	DBGPRINT1("Rp2EnableInterrupts() ERROR -  Invalid baudrate requested %d  \n", termiosBaud);
	return -1;
    }
    DBGPRINT2("Rp2SetBaudRate() - SetBaudRate port: %d baud: %d  \n", ChP->portNumber, baudRate);
    uSetBaudDivisor(ChP,  baudDivisor);
    return 0;
}

/*  
 *  Set flow control (FLOW_CONTROL_xxx) for the port
 *  Inputs:  Channel pointer and flow control setting (FLOW_CONTROL_xxx)
 */
static void Rp2SetFlowControl(CHANNEL_T *ChP, int flowControl)
{
    switch (flowControl) {
    case FLOW_CONTROL_NONE:
	DBGPRINT1("RP2: SetFlowControl to NONE on port %d  \n", ChP->portNumber);
	uDisRtsFlowCtrl(ChP);
	uDisCtsFlowCtrl(ChP);
	uDisDtrFlowCtrl(ChP);
	uDisDsrFlowCtrl(ChP);
	uDisSoftFlowCtrl(ChP);
	break;
	
    case FLOW_CONTROL_RTSCTS:
	DBGPRINT1("RP2: SetFlowControl to RTS/CTS on port %d \n", ChP->portNumber);
	uEnRtsFlowCtrl(ChP);
	uEnCtsFlowCtrl(ChP);
	uDisDtrFlowCtrl(ChP);
	uDisDsrFlowCtrl(ChP);
	uDisSoftFlowCtrl(ChP);
	break;
	
    case FLOW_CONTROL_DTRDSR:
	DBGPRINT1("RP2: SetFlowControl to DTR/DSR on port %d \n", ChP->portNumber);
	uDisRtsFlowCtrl(ChP);
	uDisCtsFlowCtrl(ChP);
	uEnDtrFlowCtrl(ChP);
	uEnDsrFlowCtrl(ChP);
	uDisSoftFlowCtrl(ChP);
	break;
	
    case FLOW_CONTROL_XONXOFF:
	DBGPRINT1("RP2: SetFlowControl to XON/XOFF on port %d \n", ChP->portNumber);
	uDisRtsFlowCtrl(ChP);
	uDisCtsFlowCtrl(ChP);
	uDisDtrFlowCtrl(ChP);
	uDisDsrFlowCtrl(ChP);
	uEnSoftFlowCtrl(ChP);
	break;
    }
}

/************************************************************************
 * GetControllerPtr - Returns a pointer to the CONTROLLER_T structure for the 
 * passed-in channel pointer.  Return NULL if error.
 ************************************************************************/
static CONTROLLER_T *Rp2GetControllerPtr(CHANNEL_T *ChP)
{
    if (ChP == NULL)
	return NULL;

    if ((ChP->portNumber >=0) && (ChP->portNumber < 32))
	return uController[0];
    else if (ChP->portNumber < 64)
	return uController[1];
    else if (ChP->portNumber < 96)
	return uController[2];
    else if (ChP->portNumber < 128)
	return uController[3];
    else
	return NULL;
}


